unit MaxInstance;

// TMaxInstance Component created in Delphi 2
//
// Author:  Michael Kelly July 27, 1999
//          mkelly99@gate.net
//
// Purpose:  Allows the Delphi programmer to set a limit on the
//           number of open instances of a Win32 application.
//
// Uses:  Win32API Semaphore calls.
//
//  Freeware:  No warranties express or implied but if you
//             use it I'd appreciate a credit or an email
//             if you find a bug.  If you give it to someone
//             else please give the whole zip file as packaged.
//
//  Status:     Still TestWare(Alpha, Beta, whatever.)
//
//  Platforms:  32 bit Windows?.  Developed and tried on
//              Win95 and just tried very quickly on the
//              Windows2000 beta.
//              I'm hoping to try it on Win98 and NT but I'm
//              encouraged that it seemed fine on Windows2000.
//              Only installed and tried with Delphi2 so far.
//
//  Installation:  See Delphi help for particulars but basically
//                 I'd recommend exploding the zip file in an
//                 empty folder in the Delphi hierarchy and saving
//                 a backup copy of your component library.  Then
//                 just follow the instructions for installing a
//                 component from source code in your version of
//                 Delphi.
//
//  Usage:  Drop on the main form of your application to limit
//          the number of instances of your app that can be run
//          concurrently.  This maximum number can be set at
//          design time using the Object Inspector.  To limit
//          the number of open instances of your App to 1 just
//          drop on the form and call the method GetSem in the
//          main form's OnCreate routine.  This minimal usage
//          is demontrated in the included TryMax program.
//
// Properties:
//
//  MaxInstances - Longint:  Max number of open App instances.
//
//  TimeOut - DWORD: milliseconds to wait for the Semaphore.
//            This number has 2 "magic cookie" values, 0,
//            meaning do not wait at all but just test if the
//            Semaphore is signalled, and -1L, meaning to wait
//            for the Semaphore forever.  Use this last with
//            caution as it can easily hang your app!  The
//            default is set to 250 or 1/4 second.
//
//  AutoKill - Boolean: If true(the default) TMaxInstance calls
//             Application.Terminate if too many apps are open.
//
//  SemName - String:  The name of the global Semaphore.  See
//            the Win32API Help for specifics but basically
//            it doesn't like a '\' directory separator char
//            in the name.  The Kernel Object namespace is all
//            one so try to use a unique name so that the first
//            instance of your App can create the Semaphore.
//
// Events:
//
//  OnSemFailure - Triggered if API Semaphore calls fail outright.
//
//  OnTooMany = Triggered if the current App is over the limit
//              of allowed instances.
//
//  Note that by setting AutoKill to False and assigning handlers
//  to the events above you can do something besides quit if too
//  many App instances are open(like chide the user or something,
//  who knows? :) or if the Semaphore cannot be created or opened.
//  For this reason I included the method FreeSem so that you can
//  free open Semaphore resources in your handlers if desired.

interface

// If doing size optimization you might be able to trim some of these
// but right now I'm more concerned on debugging the logic than experimenting
// to find out how many I can delete and still run the component.
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TMaxInstance = class(TComponent)
  private
    FMaxInstances: Longint;
    FAutoKill: Boolean;
    FSemName: string;
    FTimeOut: DWORD;
    FTooMany: TNotifyEvent;
    FSemFailure: TNotifyEvent;
  protected
    SemHandle: THandle;
    SemHeld: Boolean;
    ErrorCode: DWORD;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IsSemHeld: Boolean;
    function GetErrorCode: DWORD;
    procedure GetSem;
    procedure FreeSem;
  published
    property MaxInstances: Longint read FMaxInstances write FMaxInstances default 1;
    property AutoKill: Boolean read FAutoKill write FAutoKill default true;
    property SemName: string read FSemName write FSemName;
    property TimeOut: DWORD read FTimeOut write FTimeOut default 250;
    property OnTooMany: TNotifyEvent read FTooMany write FTooMany;
    property OnSemFailure: TNotifyEvent read FSemFailure write FSemFailure;
  end;

procedure Register;

implementation

const
  SemAccessFlags: DWORD = $000FFFFF;

procedure Register;
begin
  RegisterComponents('Freeware', [TMaxInstance]);
end;

constructor TMaxInstance.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMaxInstances := 1;
  FAutoKill := true;
  FSemName := 'MikesMaxSem';
  FTimeOut := 250;
end; // constructor

destructor TMaxInstance.Destroy;
begin
  if SemHeld then
    ReleaseSemaphore(SemHandle, 1, nil);
  if SemHandle <> 0 then
    CloseHandle(SemHandle);
  inherited Destroy;
end; // destructor

function TMaxInstance.IsSemHeld: Boolean;
begin
  Result := SemHeld;
end; // IsSemHeld

function TMaxInstance.GetErrorCode: DWORD;
begin
  Result := ErrorCode;
end; // GetErrorCode

// GetSem - The heart of the Component
//
// Any problems are likely to occur in this method where the Semaphore
// is created, opened or acquired, so I left in commented-out code to
// build and display debug msgs using ShowMessage().  If you get weird
// behavior see Delphi help for testing a component without installing
// to speed up uncommenting and recompiling or whatever if you desire.
// Note that if you uncomment the code it may unbalance the Begin..End
// pairs in statements so adjust for those changes.

procedure TMaxInstance.GetSem;
var
  WaitVal: DWORD;
  ErrorMsg, s: string;

begin
  SemHeld := false;
  ErrorCode := 0;
  SemHandle := OpenSemaphore(SemAccessFlags, BOOL(false), PChar(SemName));
  if SemHandle = 0 then
    SemHandle := CreateSemaphore(nil, MaxInstances, MaxInstances, PChar(SemName));
  if SemHandle = 0 then
  begin
    ErrorCode := GetLastError;
    if Assigned(FSemFailure) then
      OnSemFailure(Self);
    if AutoKill then
      Application.Terminate;
  end; // if SemHandle
  WaitVal := WaitForSingleObject(SemHandle, TimeOut);

  case WaitVal of
    WAIT_OBJECT_0: SemHeld := true;
    WAIT_TIMEOUT:
      begin
        if Assigned(FTooMany) then
          OnTooMany(Self);
        // ShowMessage('WaitForSingleObject Timed Out: Quiting');
        if AutoKill then
          Application.Terminate;
      end;
    WAIT_ABANDONED, WAIT_FAILED:
      begin
        ErrorCode := GetLastError;
        if Assigned(FSemFailure) then
          OnSemFailure(Self);
        // Str(ErrorCode, s);
        // ErrorMsg := 'Wait Failed: Error Code ' + s;
        // ShowMessage(ErrorMsg);
        if AutoKill then
          Application.Terminate;
      end;
  end; // case
end; // GetSem

procedure TMaxInstance.FreeSem;
begin
  ErrorCode := 0;
  if SemHeld then
    if not ReleaseSemaphore(SemHandle, 1, nil) then
    begin
      ErrorCode := GetLastError;
      exit;
    end
    else
      SemHeld := false;

  if SemHandle <> 0 then
    if not CloseHandle(SemHandle) then
      ErrorCode := GetLastError
    else
      SemHandle := 0;
end; // FreeSem

end.

